import sys
from typing import List, Sequence, Tuple, Dict
from copy import deepcopy
import numpy as np
from ase import Atoms
from ase.db import connect
from clease.settings import ClusterExpansionSettings

__all__ = ('ClusterCoverageChecker',)


class ClusterCoverageChecker:
    """
    Class for checking how large fraction of the possible configurational variation of
    all the clusters is captured by a set of template atoms. The origin of decrease in
    coverage is that for small cells, some clusters wraps around and thus interact with
    a periodic image. In a very large cell (much larger than the cluster diameters)
    there are no such "wrap-rounds".

    Example:

    In a huge cell a particular pair cluster consists of the figures
    [[0, 765], [0, 87], [0, 2], [0, 6]]

    When a small cell is used, some of the interactions will wrap around. If
    87, 765 is a periodic image of atom no. 4 and 6 is a periodic image of 2
    we have

    [[0, 4], [0, 4], [0, 2], [0, 2]]

    thus, no matter which species occupies the different sites in the small cell,
    one cannot cover all possible variation in this cluster. We define the
    coverage as <num unique figs>/<tot num figs>. Above we have two unique figs
    ([0, 4] and [0, 2]) and the total number of figures is 4. Thus, the covarage is 2/4 = 0.5.

    Some numeric examples:

    Figures                                    Coverage
    [[0, 1], [0, 5], [0, 3], [0, 7]]           1.0
    [[0, 0], [0, 0], [0, 0], [0, 0]]           0.25
    [[0, 1], [0, 1], [0, 4], [0, 8]]           0.75
    [[0, 1], [0, 1], [0, 1], [0, 3]]           0.5
    [[0, 1], [0, 1], [0, 3], [0, 3]]           0.5

    :param settings: ClusterExpansionSettings instance
    :param select_cond: Select condition used for DB query when extracing
        data from the DB. If not given, [('struct_type', '=', 'initial')]
        is used.
    """

    def __init__(self,
                 settings: ClusterExpansionSettings,
                 select_cond: Sequence[Tuple[str, str, str]] = None):
        # Make copy such that we don't alter the settings object
        self.settings = deepcopy(settings)

        if select_cond is None:
            self._select_cond = [('struct_type', '=', 'initial')]
        else:
            self._select_cond = select_cond

    def _unique_templates(self) -> List[Atoms]:
        unique_templates = []

        with connect(self.settings.db_name) as db:
            for row in db.select(self._select_cond):
                atoms = row.toatoms()
                if not _template_exists(unique_templates, atoms):
                    unique_templates.append(atoms)
        return unique_templates

    def coverage(self, template: Atoms) -> Dict[str, float]:
        """
        Return the cluster coverage for the passed atoms object.
        If the coverage is 1.0, then all possible figures in a cluster
        is covered. If it is for example 0.3, it means that the current
        template can only capture variations in 30% of the figures.

        :param template: Template
        """
        self.settings.set_active_template(atoms=template)
        coverages = {}
        for cluster in self.settings.cluster_list:
            if cluster.name.startswith(('c0', 'c1')):
                continue
            sorted_figs = []
            for figure in cluster.indices:
                sorted_figs.append(tuple(sorted(figure)))
            unique_figs = set(sorted_figs)

            coverages[cluster.name] = len(unique_figs) / len(sorted_figs)
        return coverages

    def max_coverage(self) -> Dict[str, float]:
        """
        Return the maximum cluster coverage among all templates in the database.
        If the coverage is 1.0, then all possible figures in a cluster
        is covered. If it is for example 0.3, it means that the current
        template can only capture variations in 30% of the figures.
        """
        coverages = {}
        for template in self._unique_templates():
            new_cov = self.coverage(template)
            coverages.update({k: v for k, v in new_cov.items() if v > coverages.get(k, 0.0)})
        return coverages

    def print_report(self, coverage: Dict[str, float] = None, file=sys.stdout) -> None:
        """
        Prints a nicely formatted report of coverage.

        :param coverage: dictionary with the coverage information. If not given,
            the coverage will be extracted from the database.
        :param file: a file-like object (stream); defaults to the current sys.stdout.
        """
        if coverage is None:
            coverage = self.max_coverage()

        # Add to tuple and sort from in ascending coverage order
        srt = sorted([(v, k) for k, v in coverage.items()])
        print("-" * 33, file=file)
        print("| Cluster name    | Coverage    |", file=file)
        print("-" * 33, file=file)
        for item in srt:
            print(f"| {item[1]:15} | {item[0]:8.2f}{_quality_indicator(item[0])} |", file=file)
        print("-" * 33, file=file)
        print("Explanation: ***: Insufficient, **: Poor, *: OK, '' Good", file=file)
        print(f"Cluster coverage: {_grade(coverage)}", file=file)


def _template_exists(current: Sequence[Atoms], new: Atoms) -> bool:
    """
    Check if the passed template already exists
    """
    for template in current:
        if len(template) != len(new):
            continue

        if np.allclose(template.get_positions(), new.get_positions()):
            return True
    return False


def _grade(coverage: Dict[str, float]) -> str:
    grade = "Very good (coverage > 0.75)"
    for v in coverage.values():
        if v < 0.75:
            grade = "OK (coverage > 0.5)"

        if v < 0.5:
            grade = "Insufficient (coverage < 0.5)"
            return grade
    return grade


def _quality_indicator(num: float) -> str:
    """
    Return a quality indicator. *** very bad, ** better, * even better and empty (best)
    """
    if num < 0.25:
        return '***'
    if num < 0.5:
        return '** '
    if num < 0.75:
        return '*  '
    return '   '



from typing import List, Dict, Tuple
from ase.units import kB
import numpy as np
from clease.calculator import Clease
from .mc_observer import MCObserver


class SGCState:
    """
    Represent a thermodynamic state in the semi-grand-canonical ensemble.

    :param temp: Temperature in kelvin
    :param chem_pot: Chemical potentials of the form {c1_0: -0.2, c1_1: 0.3}.
        The function `clease.tools.species_chempot2eci` is useful to convert
        chemical potentials given for each species to chemical potentials
        for each singlet.
    """

    def __init__(self, temp: float, chem_pot: Dict[str, float]):
        self.temp = temp
        self.chem_pot = chem_pot
        self.singlets = {k: 0.0 for k in self.chem_pot.keys()}
        self.avg_weight = 0.0

    def reset(self):
        self.singlets = {k: 0.0 for k in self.chem_pot.keys()}
        self.avg_weight = 0.0

    @property
    def prefix(self) -> str:
        """
        Construct a prefix based on the chemical potentials and the temperature
        """
        prefix = f"{int(self.temp)}K_"
        prefix += '_'.join(
            f"{k}{sign_indicator(v)}{int(1000.0*abs(v))}" for k, v in self.chem_pot.items())
        return prefix


def sign_indicator(v: float) -> str:
    """
    Return minus if v is negative and plus if v is positive
    """
    if v < 0.0:
        return "minus"
    return "plus"


class MultiStateSGCConcObserver(MCObserver):
    r"""
    Observer that tracks the concentration at severl different temperatures and/or chemical
    potentials. The observer utilizes the following results.
    Let :math:`A` be an observable, :math:`\beta=\frac{1}{kT}`,
    :math:`\mu` the chemical potential and :math:`n` the
    number of atoms of one of the species in a binary alloy. The average value
    of the observable is given by

    .. math::

        \langle A \rangle = \frac{\sum_{conf} A \exp{(\beta\mu n - \beta E)}}{Z(\beta, \mu)}

    where :math:`Z` is the partition function. At a different chemical potential :math:`\mu`
    and inverse temperature :math:`\beta'`,

    .. math::

        \langle A'\rangle' = \frac{\sum_{conf} A \exp{(\beta'\mu'n-\beta' E)}}{Z(\beta', \mu')}

    After some algabraic manipulation one arrives at

    .. math::

        \langle A'\rangle' =
            \frac{\langle A \exp{((\beta'\mu' - \beta\mu)n - (\beta-\beta') E)\rangle}}
            {\langle \exp{((\beta'\mu' - \beta\mu)n - (\beta-\beta') E)\rangle}}

    where the averages should be taken at inverse temperature :math:`\beta` and chemical
    potential :math:`\mu`. It should be noted that the predicted value will not be
    accurate if :math:`\mu'` or :math:`\beta'` is very different from the reference values
    :math:`\mu` and :math:`\beta`.

    :param ref_state: Reference state
    :param thermo_states: List of SGCStates where the concentration should be tracked
    :param calc: Reference to the calculator attached to the atoms object used in the
        Monte Carlo simulation
    """

    name = "multi_state_sgc_conc_observer"

    def __init__(self, ref_state: SGCState, thermo_states: List[SGCState], calc: Clease):
        super().__init__()
        self.thermo_states = thermo_states
        self.calc = calc
        self.ref_state = ref_state

    def reset(self) -> None:
        """
        Resets the observers to its initial state
        """
        for state in self.thermo_states:
            state.reset()

    def __call__(self, system_changes: List[Tuple[int, str, str]]):
        """
        Observers are called after system update, thus the calcualtor already
        reflects the changes. system_changes is therefore not used.
        """
        beta1 = 1.0 / (kB * self.ref_state.temp)
        for state in self.thermo_states:
            cf = self.calc.get_cf()
            beta2 = 1.0 / (kB * state.temp)
            betadE = sum((beta2 * state.chem_pot[k] - beta1 * self.ref_state.chem_pot[k]) * cf[k]
                         for k in state.chem_pot.keys())
            E = self.calc.energy
            weight = np.exp((beta1 - beta2) * E + betadE)
            for k in state.chem_pot.keys():
                state.singlets[k] += weight * cf[k]
                state.avg_weight += weight

    def get_averages(self) -> Dict[str, float]:
        """
        Return a dictionary with the calculated averages
        """
        averages = {}
        for state in self.thermo_states:
            for k, v in state.singlets.items():
                name = state.prefix + f"_singlet_{k}"
                averages[name] = v / state.avg_weight
        return averages



"""Definition of ClusterExpansionSettings Class.

This module defines the base-class for storing the settings for performing
Cluster Expansion in different conditions.
"""
import logging
from copy import deepcopy
from typing import List, Dict, Optional, Union, Sequence, Set, Any
from packaging.version import Version, parse

from deprecated import deprecated
import numpy as np
from ase.db import connect
from ase import Atoms

from clease.version import __version__
from clease.jsonio import jsonable
from clease.tools import wrap_and_sort_by_position
from clease.basis_function import BasisFunction
from clease.cluster import ClusterManager
from clease.basis_function import Polynomial, Trigonometric, BinaryLinear
from .concentration import Concentration
from .template_filters import ValidConcentrationFilter
from .template_atoms import TemplateAtoms
from .atoms_manager import AtomsManager

__all__ = ('ClusterExpansionSettings',)

logger = logging.getLogger(__name__)


@jsonable('ce_settings')
class ClusterExpansionSettings:
    """Base class for all Cluster Expansion settings.

    Args:
        prim (Atoms): The primitive atoms object.

        concentration (Union[Concentration, dict]): Concentration object or
            dictionary specifying the basis elements and
            concentration range of constituting species.

        size (List[int] | None, optional): Size of the supercell
            (e.g., [2, 2, 2] for 2x2x2 cell).
            ``supercell_factor`` is ignored if both ``size`` and ``supercell_factor``
            are specified. Defaults to None.

        supercell_factor (int, optional): Maximum multipilicity factor for
            limiting the size of supercell created from the primitive cell.
            ``supercell_factor`` is ignored if
            both `size` and `supercell_factor` are specified. Defaults to 27.

        db_name (str, optional): Name of the database file. Defaults to ``'clease.db'``.

        max_cluster_size (int | None, optional): Deprecated. Specifies the maximum cluster
            body size. Defaults to None. A DeprecationWarning will be raised,
            if this value is not None.

        max_cluster_dia (Sequence[float], optional): A list of int or float containing the
            maximum diameter of clusters (in Å). Defaults to ``(5., 5., 5.)``, i.e.
            a 5 Å cutoff for 2-, 3-, and 4-body clusters.

        include_background_atoms (bool, optional): Whether background elements are to
            be included. An element is considered to be a background element,
            if there is only 1 possible species which be ever be placed in a given basis.
            Defaults to False.

        basis_func_type (str, optional): Type of basis function to use. Defaults to 'polynomial'.
    """
    # pylint: disable=too-many-instance-attributes, too-many-public-methods

    # Keys which are important for saving/loading
    ARG_KEYS = ('prim_cell', 'concentration')
    KWARG_KEYS = ('size', 'supercell_factor', 'db_name', 'max_cluster_dia',
                  'include_background_atoms', 'basis_func_type')

    # Other keys we want to input back into the loaded object, but after initialization
    OTHER_KEYS = (
        'skew_threshold',
        # kwargs is a bookkeeping variable, for compatibility.
        # Just contains information about how the factory functions were called.
        'kwargs')

    def __init__(
            self,
            prim: Atoms,
            concentration: Union[Concentration, dict],
            size: Optional[List[int]] = None,
            supercell_factor: int = 27,
            db_name: str = 'clease.db',
            # max_cluster_size is only here for deprecation purposes
            # if it is not None, the user has manually specified a value
            max_cluster_size=None,
            max_cluster_dia: Sequence[float] = (5., 5., 5.),
            include_background_atoms=False,
            basis_func_type='polynomial') -> None:

        self._include_background_atoms = None
        self._cluster_mng = None
        self._trans_matrix = None
        self._cluster_list = None
        self.concentration = _get_concentration(concentration)

        self.basis_elements = deepcopy(self.concentration.basis_elements)
        self._check_first_elements()
        self.db_name = db_name
        self.size = to_3x3_matrix(size)
        self.supercell_factor = supercell_factor

        self.prim_cell = prim.copy()
        self._order_and_tag_prim_cell()
        self._store_prim_cell()

        prim_mng = AtomsManager(prim)
        prim_ind_by_basis = prim_mng.index_by_symbol([x[0] for x in self.basis_elements])
        conc_filter = ValidConcentrationFilter(concentration, prim_ind_by_basis)

        self.template_atoms = TemplateAtoms(self.prim_cell,
                                            supercell_factor=supercell_factor,
                                            size=self.size,
                                            skew_threshold=40,
                                            filters=[conc_filter])

        self.atoms_mng = AtomsManager(None)

        self.max_cluster_dia = _format_max_cluster_dia(max_cluster_dia,
                                                       max_cluster_size=max_cluster_size)

        self.set_active_template(atoms=self.template_atoms.weighted_random_template())

        self.basis_func_type = basis_func_type
        self.include_background_atoms = include_background_atoms

        if len(self.basis_elements) != self.num_basis:
            raise ValueError("list of elements is needed for each basis")

        # For storing the settings from the CLEASE factories.
        # Just for bookkeeping
        self.kwargs = {}

    @property
    def atoms(self) -> Atoms:
        return self.atoms_mng.atoms

    @property
    def max_cluster_size(self):
        return len(self.max_cluster_dia) + 1

    @property
    def all_elements(self) -> List[str]:
        return sorted([item for row in self.basis_elements for item in row])

    @property
    def num_elements(self) -> int:
        return len(self.all_elements)

    @property
    def unique_elements(self) -> List[str]:
        return sorted(list(set(deepcopy(self.all_elements))))

    @property
    def num_unique_elements(self) -> int:
        return len(self.unique_elements)

    @property
    def ref_index_trans_symm(self) -> List[int]:
        return [i[0] for i in self.index_by_sublattice]

    @property
    def skew_threshold(self):
        return self.template_atoms.skew_threshold

    @skew_threshold.setter
    def skew_threshold(self, threshold: int) -> None:
        '''
        Maximum acceptable skew level (ratio of max and min diagonal of the
        Niggli reduced cell)
        '''
        self.template_atoms.skew_threshold = threshold

    @property
    def background_indices(self) -> List[int]:
        """Get indices of the background atoms."""
        # check if any basis consists of only one element type
        basis = [i for i, b in enumerate(self.basis_elements) if len(b) == 1]

        bkg_indices = []
        for b_indx in basis:
            bkg_indices += self.index_by_basis[b_indx]
        return bkg_indices

    @property
    def cluster_mng(self):
        if self._cluster_mng is None:
            kwargs = {}
            if not self.include_background_atoms:
                kwargs['background_syms'] = self.get_bg_syms()
            self._cluster_mng = ClusterManager(self.prim_cell, **kwargs)
        return self._cluster_mng

    @property
    def include_background_atoms(self) -> bool:
        return self._include_background_atoms

    @include_background_atoms.setter
    def include_background_atoms(self, value: bool) -> None:
        if value == self._include_background_atoms:
            return
        self._include_background_atoms = value
        # We need to reset cluster manager
        self._cluster_mng = None

        self.clear_cache()
        self.basis_func_type.unique_elements = \
            self.unique_element_without_background()

    @property
    def spin_dict(self) -> Dict[str, float]:
        return self.basis_func_type.spin_dict

    @property
    def basis_functions(self):
        return self.basis_func_type.basis_functions

    @property
    def ignore_background_atoms(self) -> bool:
        return not self.include_background_atoms

    @property
    def multiplicity_factor(self) -> Dict[str, float]:
        """Return the multiplicity factor of each cluster."""
        num_sites_in_group = [len(x) for x in self.index_by_sublattice]
        return self.cluster_list.multiplicity_factors(num_sites_in_group)

    @property
    def all_cf_names(self) -> List[str]:
        num_bf = len(self.basis_functions)
        return self.cluster_list.get_all_cf_names(num_bf)

    @property
    def num_cf(self) -> int:
        """Return the number of correlation functions."""
        return len(self.all_cf_names)

    @property
    def index_by_basis(self) -> List[List[int]]:
        first_symb_in_basis = [x[0] for x in self.basis_elements]
        return self.atoms_mng.index_by_symbol(first_symb_in_basis)

    @property
    def index_by_sublattice(self) -> List[List[int]]:
        return self.atoms_mng.index_by_tag()

    @property
    def num_basis(self) -> int:
        return len(self.basis_elements)

    @property
    def basis_func_type(self):
        return self._basis_func_type

    def get_sublattice_site_ratios(self) -> np.ndarray:
        """Return the ratios of number of sites per (grouped) sublattice"""
        # Number of sites per sublattice
        sites_per_basis = np.array([len(basis) for basis in self.index_by_basis])
        num_sites = sites_per_basis.sum()
        return sites_per_basis / num_sites

    @property
    def num_active_sublattices(self) -> int:
        """Number of active sublattices"""
        return sum(self.get_active_sublattices())

    def get_active_sublattices(self) -> List[bool]:
        """List of booleans indicating if a (grouped) sublattice is active"""
        unique_no_bkg = self.unique_element_without_background()

        return [basis[0] in unique_no_bkg for basis in self.concentration.basis_elements]

    @property
    def ignored_species_and_conc(self) -> Dict[str, float]:
        """
        Return the ignored species and their concentrations normalised to the total number
        of atoms.
        """
        unique_no_bkg = self.unique_element_without_background()
        # Find the concentration within grouped sublattices
        orig_basis = self.concentration.basis_elements
        # Concentration of sites within each sublattice
        ratios = self.get_sublattice_site_ratios()
        assert len(orig_basis) == len(ratios)
        ignored = {}
        for ratio, basis in zip(ratios, orig_basis):
            elem = basis[0]
            if elem not in unique_no_bkg:
                if len(basis) != 1:
                    raise ValueError(("Ignored sublattice contains multiple elements -"
                                      "this does not make any sense"))
                if elem not in ignored:
                    ignored[elem] = ratio
                else:
                    # This element is already on one of the ignored background here we
                    # accumulate the concentration
                    ignored[elem] += ratio
        return ignored

    @property
    def atomic_concentration_ratio(self) -> float:
        """
        Ratio between true concentration (normalised to atoms) and the internal concentration used.
        For example, if one of the two basis is fully occupied, and hence ignored internally, the
        internal concentration is half of the actual atomic concentration.
        """
        ratios = self.get_sublattice_site_ratios()
        # We only want to include active sublattices
        active_sublatt = self.get_active_sublattices()
        # Add up all ratios for the active sublattices
        return ratios[active_sublatt].sum()

    @basis_func_type.setter
    def basis_func_type(self, bf_type):
        """
        Type of basis function to use.
        It should be one of "polynomial", "trigonometric" or "binary_linear"
        """
        unique_element = self.unique_element_without_background()

        if isinstance(bf_type, BasisFunction):
            if bf_type.unique_elements != sorted(unique_element):
                raise ValueError("Unique elements in BasisFunction instance "
                                 "is different from the one in settings")
            self._basis_func_type = bf_type
        elif isinstance(bf_type, str):
            if bf_type.lower() == 'polynomial':
                self._basis_func_type = Polynomial(unique_element)
            elif bf_type.lower() == 'trigonometric':
                self._basis_func_type = Trigonometric(unique_element)
            elif bf_type.lower() == "binary_linear":
                self._basis_func_type = BinaryLinear(unique_element)
            else:
                msg = f"basis function type {bf_type} is not supported."
                raise ValueError(msg)
        else:
            raise ValueError("basis_function has to be an instance of BasisFunction or a string")

    def get_bg_syms(self) -> Set[str]:
        """
        Return the symbols in the basis where there is only one element
        """
        return set(x[0] for x in self.basis_elements if len(x) == 1)

    def unique_element_without_background(self):
        """Remove background elements."""
        if self.include_background_atoms:
            bg_sym = set()
        else:
            bg_sym = self.get_bg_syms()

            # Remove bg_syms that are also present in basis with more than one
            # element
            for elems in self.basis_elements:
                if len(elems) == 1:
                    continue
                to_be_removed = set()
                for s in bg_sym:
                    if s in elems:
                        to_be_removed.add(s)

                bg_sym -= to_be_removed

        unique_elem = set()
        for x in self.basis_elements:
            unique_elem.update(x)
        return list(unique_elem - bg_sym)

    def prepare_new_active_template(self, template):
        """Prepare necessary data structures when setting new template."""
        logger.debug('Preparing new template in settings')
        self.atoms_mng.atoms = template
        self.clear_cache()

    def set_active_template(self, atoms=None):
        """Set a new template atoms object."""
        if atoms is not None:
            template = self.template_atoms.get_template_matching_atoms(atoms=atoms)
        else:
            template = self.template_atoms.weighted_random_template()

        template = wrap_and_sort_by_position(template)

        if atoms is not None:
            # Check that the positions of the generated template
            # matches the ones in the passed object
            atoms = wrap_and_sort_by_position(atoms)
            if not np.allclose(template.get_positions(), atoms.get_positions()):
                raise ValueError(f"Inconsistent positions. Passed object\n"
                                 f"{atoms.get_positions()}\nGenerated template"
                                 f"\n{template.get_positions()}")
        self.prepare_new_active_template(template)

    def _order_and_tag_prim_cell(self):
        """
        Add a tag to all the atoms in the unit cell to track the sublattice.
        Tags are added such that that the lowest tags corresponds to "active"
        sites, while the highest tags corresponds to background sites. An
        example is a system having three sublattices that can be occupied by
        more than one species, and two sublattices that can be occupied by
        only one species. The tags 0, 1, 2 will then be assigned to the three
        sublattices that can be occupied by more than one species, and the two
        remaining lattices will get the tag 3, 4.

        Re-orders the primitive cell in accordance to the order of the tags.
        """
        bg_sym = self.get_bg_syms()
        tag = 0

        # Tag non-background elements first
        for atom in self.prim_cell:
            if atom.symbol not in bg_sym:
                atom.tag = tag
                tag += 1

        # Tag background elements
        for atom in self.prim_cell:
            if atom.symbol in bg_sym:
                atom.tag = tag
                tag += 1

        # Rearange primitive cell in order of the tags.
        sorted_indx = np.argsort([atom.tag for atom in self.prim_cell])
        self.prim_cell = self.prim_cell[sorted_indx]

    def _store_prim_cell(self):
        """Store unit cell to the database. Returns the id of primitive cell in the database"""
        with connect(self.db_name) as db:
            shape = self.prim_cell.cell.cellpar()
            for row in db.select(name='primitive_cell'):
                uc_shape = row.toatoms().cell.cellpar()
                if np.allclose(shape, uc_shape):
                    return row.id

            uid = db.write(self.prim_cell, name='primitive_cell')
        return uid  # Ensure connection is closed before returning

    def _get_prim_cell(self):
        raise NotImplementedError("This function has to be implemented in in derived classes.")

    @property
    def trans_matrix(self):
        """Get the translation matrix, will be created upon request"""
        self.ensure_clusters_exist()
        return self._trans_matrix

    @property
    def cluster_list(self):
        """Get the cluster list, will be created upon request"""
        self.ensure_clusters_exist()
        return self._cluster_list

    def clear_cache(self) -> None:
        """Clear the cached objects, due to a change e.g. in the template atoms"""
        logger.debug('Clearing the cache')
        self._trans_matrix = None
        self._cluster_list = None

    def create_cluster_list_and_trans_matrix(self):
        """Prepares the internal cache objects by calculating cluster related properties"""
        logger.debug('Creating translation matrix and cluster list')
        at_cpy = self.atoms.copy()
        self.cluster_mng.build(self.max_cluster_dia)
        self._cluster_list = self.cluster_mng.info_for_template(at_cpy)
        self._trans_matrix = self.cluster_mng.translation_matrix(at_cpy)

    def view_clusters(self) -> None:
        """Display all clusters along with their names."""
        # pylint: disable=import-outside-toplevel
        from ase.gui.gui import GUI
        from ase.gui.images import Images
        figures = self.get_all_figures_as_atoms()
        images = Images()
        images.initialize(figures)
        gui = GUI(images, expr='')
        gui.show_name = True
        gui.run()

    def get_all_figures_as_atoms(self) -> List[Atoms]:
        """Get the list of all possible figures, in their
        ASE Atoms representation."""
        self.ensure_clusters_exist()
        return self.cluster_mng.get_figures()

    def ensure_clusters_exist(self) -> None:
        """Ensure the cluster list and trans matrix has been populated.
        They are not calculated upon creaton of the settings instance,
        for performance reasons. They will be constructed if required.
        Nothing is done if the cache exists."""
        if self._cluster_list is None or self._trans_matrix is None:
            logger.debug('Triggered construction of clusters')
            self.create_cluster_list_and_trans_matrix()
            # It should not be possible for them to be None after this call.
            assert self._cluster_list is not None
            assert self._trans_matrix is not None

    def get_all_templates(self):
        """
        Return a list with all template atoms
        """
        return self.template_atoms.get_all_templates()

    def view_templates(self):
        """
        Display all templates in the ASE GUi
        """
        # pylint: disable=import-outside-toplevel
        from ase.visualize import view
        view(self.get_all_templates())

    def _check_first_elements(self):
        basis_elements = self.basis_elements
        num_basis = self.num_basis
        # This condition can be relaxed in the future
        first_elements = []
        for elements in basis_elements:
            first_elements.append(elements[0])
        if len(set(first_elements)) != num_basis:
            raise ValueError("First element of different basis should not be the same.")

    def todict(self) -> Dict:
        """Return a dictionary representation of the settings class.

        Example:

            >>> from clease.settings import CEBulk, Concentration
            >>> conc = Concentration([['Au', 'Cu']])
            >>> settings = CEBulk(conc, crystalstructure='fcc', a=4.1)
            >>> dct = settings.todict()  # Get the dictionary representation
        """
        vars_to_save = self.ARG_KEYS + self.KWARG_KEYS + self.OTHER_KEYS
        dct = {'clease_version': str(__version__)}
        for key in vars_to_save:
            val = getattr(self, key)
            dct[key] = val
        return dct

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> 'ClusterExpansionSettings':
        """Load a new ClusterExpansionSettings class from a dictionary representation.

        Example:

            >>> from clease.settings import CEBulk, Concentration, ClusterExpansionSettings
            >>> conc = Concentration([['Au', 'Cu']])
            >>> settings = CEBulk(conc, crystalstructure='fcc', a=4.1)
            >>> dct = settings.todict()  # Get the dictionary representation
            >>> # Remove the existing settings, perhaps due to being in a new environment
            >>> del settings
            >>> # Load in the settins from the dictionary representation
            >>> settings = ClusterExpansionSettings.from_dict(dct)
        """
        dct = deepcopy(dct)
        version = dct.pop('clease_version', None)
        if version is None or parse(version) < Version('0.10.2'):
            # For backwards compatibility
            # pylint: disable=import-outside-toplevel, cyclic-import
            from .utils import old_settings_from_json
            return old_settings_from_json(dct)

        # Get the args and kwargs we expect in our function signature
        args = [dct.pop(key) for key in cls.ARG_KEYS]

        # Should we allow for missing kwargs keys?
        kwargs = {key: dct.pop(key) for key in cls.KWARG_KEYS}

        if 'max_cluster_size' in dct:
            # For compatibility, since it's no longer in KWARG_KEYS
            kwargs['max_cluster_size'] = dct.pop('max_cluster_size')

        settings = cls(*args, **kwargs)

        for key in cls.OTHER_KEYS:
            # Populate "other" keys, which aren't *args or **kwargs
            value = dct.pop(key)
            setattr(settings, key, value)

        if dct:
            # We have some unexpected left-overs
            logger.warning('Unused items from dictionary: %s', dct)
        return settings

    def clusters_table(self) -> str:
        """String with information about the clusters"""
        mult_dict = self.multiplicity_factor

        columns = ['Index', 'Cluster Name', 'Size', 'Group', 'Radius', 'Figures', 'Multiplicity']

        fmt = '| {:<5} | {:<12} | {:<4} | {:<5} | {:<6} | {:<7} | {:<12} |'
        header = fmt.format(*columns)
        rule = '-' * len(header)  # Horizontal line of -----
        lines = [rule, header, rule]

        for ii, cluster in enumerate(self.cluster_list.clusters):
            name = cluster.name
            size = f'{cluster.size:d}'.center(4, ' ')
            mult = f'{int(mult_dict[name]):d}'.center(12, ' ')
            radius = f'{cluster.diameter / 2:2.4f}'
            n_figures = f'{len(cluster.figures)}'.center(7)
            group = f'{cluster.group:d}'.center(5)
            index_s = f'{ii:d}'.center(5)

            s = fmt.format(index_s, name, size, group, radius, n_figures, mult)
            lines.append(s)

        lines.append(rule)

        return '\n'.join(lines)


def to_3x3_matrix(size):
    if size is None:
        return None
    if isinstance(size, np.ndarray):
        size = size.tolist()

    # Is already a matrix
    if np.array(size).shape == (3, 3):
        return size

    if np.array(size).shape == (3,):
        return np.diag(size).tolist()

    raise ValueError(f"Cannot convert passed array {size} to 3x3 matrix")


def _get_concentration(concentration: Union[Concentration, dict]) -> Concentration:
    """Helper function to format the concentration"""
    if isinstance(concentration, Concentration):
        conc = concentration
    elif isinstance(concentration, dict):
        conc = Concentration.from_dict(concentration)
    else:
        raise TypeError("concentration has to be either dict or instance of Concentration")
    return conc


def _format_max_cluster_dia(max_cluster_dia, max_cluster_size=None):
    """Formatter of max_cluster_dia."""
    if max_cluster_size is None and not isinstance(max_cluster_dia, (int, float)):
        # Assume max_cluster_dia is sequence[float], and user didn't specify any
        # (now deprecated) max_cluster sizes.
        return np.array(max_cluster_dia)
    return _old_format_max_cluster_dia(max_cluster_dia, max_cluster_size)


def _old_format_max_cluster_dia(max_cluster_dia, max_cluster_size):

    # User specified an old version of MCS and MCD
    dep_msg = f"""
    max_cluser_size should no longer be specfied explicitly,
    and max_cluster_dia should no longer be an int or float.

    Specify cluster sizes with max_cluster_dia as an array-like instead.
    Got max_cluster_size '{max_cluster_size}' and max_cluster_dia '{max_cluster_dia}'.
    Try instead to use max_cluster_dia as an array, e.g. max_cluster_dia=[5., 5., 5.]
    for 2-, 3- and 4-body clusters of cutoff 5 Å.
    """

    @deprecated(version='0.10.6', reason=dep_msg)
    def _formatter():
        # max_cluster_dia is list or array
        if isinstance(max_cluster_dia, (list, np.ndarray, tuple)):
            # Length should be either max_cluster_size+1 or max_cluster_size-1
            mcd = np.array(max_cluster_dia, dtype=float)
            if len(max_cluster_dia) == max_cluster_size + 1:
                # Remove the first two entries, assume they are 0- and 1-body diameters
                mcd = mcd[2:]
            elif len(max_cluster_dia) == max_cluster_size - 1:
                # Assume max_cluster_dia contains 2+ body clusters
                pass
            else:
                raise ValueError("Invalid length for max_cluster_dia.")
        elif isinstance(max_cluster_dia, (int, float)):
            if max_cluster_size is None:
                raise ValueError('Received no max_cluster_size, but a float for max_cluster_dia')
            mcd = np.ones(max_cluster_size - 1, dtype=float) * max_cluster_dia
        # Case for *None* or something else
        else:
            raise TypeError(f"max_cluster_dia is of wrong type, got: {type(max_cluster_dia)}")
        return mcd.round(decimals=3)

    return _formatter()



"""Import the C++ definition of an Atoms object to Cython"""
# distutils: language = c++

from libcpp.vector cimport vector
from libcpp.string cimport string

from four_vector cimport FourVector

cdef extern from "atoms.hpp":
    cdef cppclass Atoms:
        Atoms(object atoms, object four_vectors) except +

        int Ns, Nx, Ny, Nz

        vector[int] get_numbers() const
        vector[string] get_symbols() const

        vector[FourVector] get_four_vectors() const

        int get_1d_index(FourVector vec) const

        void apply_change(object single_change)
        void undo_change(object single_change)



from pathlib import Path
from packaging.version import parse

with Path(__file__).with_name('_version.txt').open('r') as f:
    __version__ = parse(f.readline().strip())

# Representation of version_info as (x, y, z)
version_info = __version__.release

__all__ = ('__version__', 'version_info')



import time
import json
import logging
from itertools import product
import numpy as np
from scipy.special import polygamma
from scipy.optimize import brentq
from matplotlib import pyplot as plt
from .regression import LinearRegression

logger = logging.getLogger(__name__)

__all__ = ('BayesianCompressiveSensing',)


class BayesianCompressiveSensing(LinearRegression):
    """
    Fit a sparse CE model to data. Based on the method
    described in

    Babacan, S. Derin, Rafael Molina, and Aggelos K. Katsaggelos.
    "Bayesian compressive sensing using Laplace priors."
    IEEE Transactions on Image Processing 19.1 (2010): 53-63.

    Different values has different priors.

    1. For the ECIs a normal distribution is assumed
        (the i-th eci is: eci_i -- N(J | 0, var_i)=
    2. The inverce variance of each ECI is gamma distributed
        (i.e. 1/var_i -- gamma(x | 1, lambda/2))
    3. The lambda parameter above is also gamma distributed
        (i.e. lamb -- gamma(x | shape_lamb/2, shape_lamb/2))
    4. The noise parameter is uniformly distributed on the
        positive axis (i.e. noise -- uniform(x | 0, inf)

    Parameters:

    shape_var: float
        Shape parameter for the gamma distribution for the
        inverse variance (1/var -- gamma(x | shape_var/2, rate_var/2))
    rate_var: float
        Rate parameter for the gamma distribution for the
        inverse variance (1/var -- gamma(x | shape_var/2, rate_var/2))
    shape_lamb: float
        Shape parameter for gamma distribution for the
        lambda parameter (lambda -- gamma(x | 1, shape_lamb))
    variance_opt_start: int
        Optimization of inverse variance starts after this amount
        of iterations
    lamb_opt_start: int
        Optimization of lambda and shape_lamb starts after this
        amount of iterations. If this number is set very high,
        lambda will be kept at zero, making the algorithm
        efficitively a Relvance Vector Machine (RVM)
    fname: str
        Backup file for parameters
    maxiter: int
        Maximum number of iterations
    output_rate_sec: int
        Interval in seconds between status messages
    select_strategy: str
        Strategy for selecting new correlation function for each iteration.
        If 'max_increase' it will select the basis function that leads
        to the largest increase in likelihood value.
        If 'random' correlation functions are selected at random
    noise: float
        Initial estimate of the noise in the data
    init_lamb: float
        Initial value for the lambda parameter
    penalty: float
        Penalization value added to the diagonal of matrice
        to avoid singular matrices
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self,
                 shape_var=0.5,
                 rate_var=0.5,
                 shape_lamb=0.5,
                 lamb_opt_start=200,
                 variance_opt_start=100,
                 fname="bayes_compr_sens.json",
                 maxiter=100000,
                 output_rate_sec=2,
                 select_strategy="max_increase",
                 noise=0.1,
                 init_lamb=0.0,
                 penalty=1E-8):
        # pylint: disable=too-many-arguments
        super().__init__()

        # Paramters
        self.shape_var = shape_var
        self.rate_var = rate_var
        self.shape_lamb = shape_lamb
        self.variance_opt_start = variance_opt_start
        self.maxiter = maxiter
        self.output_rate_sec = output_rate_sec
        self.select_strategy = select_strategy
        self.fname = fname
        self.noise = noise
        self.lamb_opt_start = lamb_opt_start
        self.penalty = penalty

        # Store a copy of all the key-word arguments
        # passed by the user
        self.user_supplied_args = {
            "shape_var": shape_var,
            "rate_var": rate_var,
            "shape_lamb": shape_lamb,
            "lamb_opt_start": lamb_opt_start,
            "variance_opt_start": variance_opt_start,
            "fname": fname,
            "maxiter": maxiter,
            "output_rate_sec": output_rate_sec,
            "select_strategy": select_strategy,
            "noise": noise,
            "init_lamb": init_lamb,
            "penalty": penalty
        }

        # Arrays used during fitting
        self.X = None
        self.y = None

        self.gammas = None
        self.eci = None
        self.inv_variance = None
        self.lamb = init_lamb
        self.inverse_sigma = None
        self.eci = None

        # Quantities used for fast updates
        self.S = None
        self.Q = None
        self.ss = None
        self.qq = None

    def _initialize(self, reset=False):
        """
        Initialize all parameters after X and y is given
        """
        if reset:
            self.__dict__.update(self.user_supplied_args)
        num_features = self.X.shape[1]
        if reset or self.gammas is None:
            self.gammas = np.zeros(num_features)
        self.eci = np.zeros_like(self.gammas)

        if reset or self.inv_variance is None:
            self.inv_variance = 1.0 / self.noise**2

        if reset or self.lamb is None:
            self.lamb = 0.0

        self.inverse_sigma = np.zeros((num_features, num_features))
        self.eci = np.zeros(num_features)

        if self.X is not None:
            # Quantities used for fast updates
            self.S = np.diag(self.inv_variance * self.X.T.dot(self.X))
            self.Q = self.inv_variance * self.X.T.dot(self.y)
            self.ss = self.S / (1.0 - self.gammas * self.S)
            self.qq = self.Q / (1.0 - self.gammas * self.S)

    def precision_matrix(self, X):
        """
        Return the precision matrix needed by the Evaluate class.
        Only contributions from the correlation functions with
        gamma > 0 are included.
        """
        if not np.allclose(X, self.X):
            raise RuntimeError("Inconsistent design matrix given!")
        sel_indx = self.selected
        X_sel = self.X[:, sel_indx]
        N = len(sel_indx)
        prec = np.linalg.inv(X_sel.T.dot(X_sel) + self.penalty * np.eye(N))

        N = self.X.shape[1]
        full_prec = np.zeros((N, N))

        indx = list(range(len(sel_indx)))
        for i in product(indx, indx):
            full_prec[sel_indx[i[0]], sel_indx[i[1]]] = prec[i[0], i[1]]
        return full_prec

    def mu(self):
        """
        Calculate the expectation value for the ECIs
        """
        sel = self.selected
        return self.inv_variance * self.inverse_sigma.dot(self.X[:, sel].T.dot(self.y))

    def optimal_gamma(self, indx):
        """
        Return the gamma value that maximize the likelihood

        Parameters:

        indx: int
            Index of the selected correlation function
        """
        s = self.ss[indx]

        # Conda had divide by zero error
        if abs(s) < 1E-6:
            s = 1E-6
        qsq = self.qq[indx]**2

        if self.lamb < 1E-6:
            return (self.qq[indx] / s)**2 - 1.0 / s

        term1 = s + 2 * self.lamb

        delta = s**2 + 4 * self.lamb * qsq

        gamma = (np.sqrt(delta) - term1) / (2 * self.lamb * s)
        return gamma

    def optimal_lamb(self):
        """Calculate the optimal value for the lambda parameter. """
        N = self.X.shape[1]  # Number of ECIs
        return 2*(N - 1 + 0.5*self.shape_lamb) / \
            (np.sum(self.gammas) + self.shape_lamb)

    def optimal_inv_variance(self):
        """Calculate the optimal value for the inverse variance"""
        N = self.X.shape[0]  # Number of data points
        a = 1.0
        b = 0.0
        mse = np.sum((self.y - self.X.dot(self.eci))**2)
        return (0.5 * N + a) / (0.5 * mse + b)

    def optimal_shape_lamb(self):
        """Calculate the optimal value for the shape paremeter for lambda."""
        res = brentq(shape_parameter_equation, 1E-30, 1E100, args=(self.lamb,), maxiter=10000)
        return res

    def update_quantities(self):
        """Update helper parameters needed for the next iteration."""
        sel = self.selected
        X_sel = self.X[:, sel]
        prec = X_sel.dot(self.inverse_sigma).dot(X_sel.T)

        self.S = \
            np.diag(self.inv_variance*self.X.T.dot(self.X)
                    - self.inv_variance**2 * self.X.T.dot(prec.dot(self.X)))
        self.Q = self.inv_variance*self.X.T.dot(self.y) - \
            self.inv_variance**2 * self.X.T.dot(prec.dot(self.y))

        self.ss = self.S / (1.0 - self.gammas * self.S)
        self.qq = self.Q / (1.0 - self.gammas * self.S)

    @property
    def selected(self):
        return np.argwhere(self.gammas > 0.0)[:, 0]

    def update_sigma_mu(self):
        """Update sigma and mu."""
        X_sel = self.X[:, self.selected]
        sigma = self.inv_variance*X_sel.T.dot(X_sel) + \
            np.diag(1.0/self.gammas[self.selected])
        try:
            self.inverse_sigma = np.linalg.inv(sigma)
        except np.linalg.LinAlgError:
            self.inverse_sigma = np.linalg.pinv(sigma)
        self.eci[self.selected] = self.mu()

    def get_basis_function_index(self, select_strategy) -> int:
        """Select a new correlation function."""
        if select_strategy == "random":
            return np.random.randint(low=0, high=len(self.gammas))
        if select_strategy == "max_increase":
            return self._get_bf_with_max_increase()
        raise ValueError(f'Unknown select strategy: {select_strategy}')

    def log_likelihood_for_each_gamma(self, gammas):
        """Log likelihood value for all gammas.

        Arguments
        =========
        gammas: np.ndarray
            Value for all the gammas
        """
        denum = 1 + gammas * self.ss

        # quantities leading to a negativie denuminator should not
        # be possible (but because of iterative increment they occure)
        # once in a while. Set such numbers to a number close to zero.
        denum[denum <= 0.0] = 1E-10
        return np.log(1 / denum) + self.qq**2 * gammas / denum - self.lamb * gammas

    def _get_bf_with_max_increase(self):
        """Return the index of the correlation function that leads
           to the largest increase in likelihiood value."""

        new_gammas = np.array([self.optimal_gamma(i) for i in range(len(self.gammas))])

        if np.all(new_gammas < 0.0):
            rand_gam = np.random.randint(0, high=(len(self.gammas)))
            logger.warning(
                "Cannot determine which gamma to include. Trying random selection, gamma=%.3f ...",
                rand_gam)
            return rand_gam

        new_gammas[new_gammas < 0.0] = 0.0
        current_likeli = self.log_likelihood_for_each_gamma(self.gammas)
        new_likeli = self.log_likelihood_for_each_gamma(new_gammas)

        diff = new_likeli - current_likeli
        return np.argmax(diff)

    def rmse(self):
        """Return root mean square error."""
        indx = self.selected
        pred = self.X[:, indx].dot(self.eci[indx])
        return np.sqrt(np.mean((pred - self.y)**2))

    @property
    def num_ecis(self):
        return np.count_nonzero(self.gammas)

    def todict(self):
        """Convert all parameters to a dictionary."""
        data = {}

        # Selection of vars we want to save
        vars_to_save = ("inv_variance", 'gammas', 'shape_var', 'rate_var', 'shape_lamb', 'lamb',
                        'maxiter', 'output_rate_sec', 'select_strategy', 'noise', 'lamb_opt_start')
        for var in vars_to_save:
            val = getattr(self, var)
            if isinstance(val, np.ndarray):
                val = val.tolist()
            data[var] = val

        return data

    def save(self):
        """Save the results from file."""
        with open(self.fname, 'w') as outfile:
            json.dump(self.todict(), outfile)
        logger.info("Backup data written to %s", self.fname)

    @staticmethod
    def load(fname):
        bayes = BayesianCompressiveSensing()
        bayes.fname = fname
        with open(fname, 'r') as infile:
            data = json.load(infile)

        for key, value in data.items():
            if value is not None:
                setattr(bayes, key, value)
        return bayes

    def __eq__(self, other):
        """Compare to BayesianCompressiveSensing objects."""
        equal = True

        # Required fields to be equal if two objects
        # should be considered equal
        items = [
            "fname", "gammas", "inv_variance", "lamb", "shape_var", "rate_var", "shape_lamb",
            "lamb", "maxiter", "select_strategy", "output_rate_sec", "noise", "variance_opt_start"
        ]
        for k in items:
            v = self.__dict__[k]
            if isinstance(v, np.ndarray):
                equal = equal and np.allclose(v, other.__dict__[k])
            elif isinstance(v, float):
                equal = equal and abs(v - other.__dict__[k]) < 1E-6
            else:
                equal = equal and (v == other.__dict__[k])
        return equal

    def estimate_loocv(self):
        """Return an estimate of the LOOCV."""
        X_sel = self.X[:, self.selected]
        e_pred = self.X.dot(self.eci)
        delta_e = e_pred - self.y
        N = X_sel.shape[1]
        prec = np.linalg.inv(X_sel.T.dot(X_sel) + self.penalty * np.eye(N))

        delta_cv = delta_e / (1 - np.diag(X_sel.dot(prec).dot(X_sel.T)))
        cv_sq = np.mean(delta_cv**2)
        return np.sqrt(cv_sq)

    def fit(self, X, y):
        """Fit ECIs to the data

        Parameters:

        X: np.ndarray
            Design matrix (NxM: N number of datapoints,
            M number of correlation functions)
        y: np.ndarray
            Array of length N with the energies
        """
        # pylint: disable=too-many-branches, too-many-statements
        # XXX: Needs some cleaning
        allowed_strategies = ["random", "max_increase"]

        if self.select_strategy not in allowed_strategies:
            msg = f"select_strategy has to be one of {allowed_strategies}"
            raise ValueError(msg)

        self.X = X
        self.y = y
        self._initialize(reset=True)

        is_first = True
        iteration = 0
        now = time.time()
        d_gamma = 1E100
        while iteration < self.maxiter:
            if time.time() - now > self.output_rate_sec:
                msg = f"Iter: {iteration} "
                msg += f"RMSE: {1000.0*self.rmse():.3E} "
                msg += f"LOOCV (approx.): {1000.0*self.estimate_loocv():.3E} "
                msg += f"Num ECI: {self.num_ecis} "
                msg += f"Lamb: {self.lamb:.3E} "
                msg += f"Shape lamb: {self.shape_lamb:.3E} "
                msg += f"Noise: {np.sqrt(1.0/self.inv_variance):.3E}"
                logger.info(msg)
                now = time.time()

            iteration += 1
            already_excluded = False

            if is_first:
                indx = np.argmax(self.qq**2 - self.ss)
                is_first = False
            else:
                indx = self.get_basis_function_index(self.select_strategy)
            gamma = self.optimal_gamma(indx)

            d_gamma = 0.0
            if gamma > 0.0:
                d_gamma = gamma - self.gammas[indx]
                self.gammas[indx] = gamma
            else:
                gamma = self.gammas[indx]

                if abs(gamma) < 1E-6:
                    already_excluded = True

                self.gammas[indx] = 0.0
                self.eci[indx] = 0.0

            if already_excluded:
                continue

            self.update_sigma_mu()
            self.update_quantities()

            if iteration > self.lamb_opt_start:
                self.lamb = self.optimal_lamb()
                self.shape_lamb = self.optimal_shape_lamb()

            if iteration > self.variance_opt_start:
                self.inv_variance = self.optimal_inv_variance()

            if abs(d_gamma) < 1E-8:
                break

        # Save backup for future restart
        if self.fname:
            self.save()

        return self.eci

    def show_shape_parameter(self):
        """Show a plot of the transient equation for the optimal
            shape parameter for lambda."""
        x = np.logspace(-10, 10)

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(x, shape_parameter_equation(x, self.lamb))
        ax.axhline(0, ls="--")
        ax.set_xscale("log")
        plt.show()

    @property
    def weight_matrix(self):
        return LinearRegression.weight_matrix(self)


def shape_parameter_equation(x, lamb):
    return np.log(x / 2.0) + 1 - polygamma(0, x / 2) + np.log(lamb) - lamb



# pylint: disable=import-outside-toplevel
from tkinter import TclError
from matplotlib import pyplot as plt
from ase.db import connect
from ase.gui.gui import GUI
from ase.gui.images import Images


class InteractivePlot:
    """Class for creating an interactive plot with annotations

    Parameters:

    fig: Figure instance
        Instance of the figure object visualizing the data

    ax: Axes instance
        Instance of the axes object containing the lines

    lines: Array of Line objects that annotations apply to

    annotations: Nested list with annotations
        Each list contains one annotation for each (x, y) pair on that
        line
    """

    def __init__(self, fig, ax, lines, annotations):
        self.fig = fig
        self.ax = ax
        self.lines = lines
        self.annotations = annotations
        self.active_annot = ax.annotate("",
                                        xy=(0, 0),
                                        xytext=(-20, 20),
                                        textcoords="offset points",
                                        bbox=dict(boxstyle="round", fc="w"),
                                        arrowprops=dict(arrowstyle="->"))
        self.active_annot.set_visible(False)
        self.active_line_index = 0

        for i, line in enumerate(self.lines):
            x, _ = line.get_data()
            if len(x) != len(self.annotations[i]):
                msg = f"Annotations for line {i} "
                msg += f"has length {len(self.annotations[i])} "
                msg += f"but there are {len(x)} data points!"
                raise ValueError(msg)

        self.fig.canvas.mpl_connect("motion_notify_event", self.hover)
        plt.show()

    def _update_annotation(self, ind):
        """Update the annotation shown."""
        try:
            line = self.lines[self.active_line_index]
            x, y = line.get_data()
            self.active_annot.xy = (x[ind["ind"][0]], y[ind["ind"][0]])
            anot = self.annotations[self.active_line_index]
            text = anot[ind["ind"][0]]
            self.active_annot.set_text(text)
        except IndexError:
            pass

    def hover(self, event):
        """React on a hover event."""
        vis = self.active_annot.get_visible()
        if event.inaxes == self.ax:
            cont = False
            for i, line in enumerate(self.lines):
                cont, ind = line.contains(event)
                if cont:
                    self.active_line_index = i
                    break

            if cont:
                self._update_annotation(ind)
                self.active_annot.set_visible(True)
                self.fig.canvas.draw_idle()
            else:
                if vis:
                    self.active_annot.set_visible(False)
                    self.fig.canvas.draw_idle()


class ShowStructureOnClick(InteractivePlot):

    def __init__(self, fig, ax, lines, names, db_name):
        self.db_name = db_name
        self.active_images = Images()
        # Uninitilized GUI. We don't create this until we click,
        # since we otherwise just open a new empty GUI window.
        self.gui = None
        fig.canvas.mpl_connect("button_press_event", self.on_click)
        super().__init__(fig, ax, lines, names)

    def on_click(self, event):

        if event.inaxes != self.ax:
            return

        if event.button == 1:
            if self.gui is not None:
                # Try to close the existing GUI, if we already opened one
                try:
                    self.gui.exit()
                except TclError:
                    pass
            db = connect(self.db_name)
            atoms = []

            # Find the index of the point
            for i, line in enumerate(self.lines):
                cont, ind = line.contains(event)
                if cont:
                    self.active_line_index = i
                    break

            name = self.annotations[self.active_line_index][ind["ind"][0]]
            for row in db.select(name=name):
                atoms.append(row.toatoms())
            self.active_images.initialize(atoms)
            # Create a new GUI instance and run it
            self.gui = GUI(self.active_images)
            self.gui.run()



"""Deprecated import of new_struct.
Module has been moved to clease.structgen, so
use clease.structgen.new_struct instead """

from warnings import warn
from deprecated import deprecated
import numpy as np
from clease.structgen import new_struct

MSG = 'Import {} from clease.structgen.new_struct instead'
DEP_VERSION = '0.10.2'  # Deprecation version

MODULE_DEP_MSG = f"""
The clease.new_struct module has been moved as of version {DEP_VERSION}.
Please use the clease.structgen.new_struct module instead"""

# Print the message on import
warn(MODULE_DEP_MSG, np.VisibleDeprecationWarning)


@deprecated(version=DEP_VERSION, reason=MSG.format('NewStructures'))
class NewStructures(new_struct.NewStructures):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('MaxAttemptReachedError'))
class MaxAttemptReachedError(new_struct.MaxAttemptReachedError):
    pass



import os
from distutils.sysconfig import get_python_inc
from setuptools import setup, Extension


def src_folder():
    candidates = ['cxx/src/', 'clease/', './']
    for c in candidates:
        if os.path.exists(c + 'cluster.cpp'):
            return c
    raise RuntimeError("Cannot find source folder.")


def get_cython_folder():
    candidates = ['cxx/cython/', 'cython/', './']
    for c in candidates:
        if os.path.exists(c + 'clease_cxx.pyx'):
            return c
    raise RuntimeError("Cannot find cython folder")


def include_folder():
    candidates = ['cxx/include/', 'clease/', './']
    for c in candidates:
        if os.path.exists(c + 'cluster.hpp'):
            return c
    raise RuntimeError("Cannot find include folder")


def get_npy_include_folder():
    import numpy as np
    return np.get_include()


def build_ext(ext_module):
    from Cython.Build import cythonize
    return cythonize(ext_module, compiler_directives={"language_level": "3"})


cxx_src_folder = src_folder()
cxx_inc_folder = include_folder()
cython_folder = get_cython_folder()

src_files = [
    'cf_history_tracker.cpp',
    'additional_tools.cpp',
    'cluster.cpp',
    'row_sparse_struct_matrix.cpp',
    'named_array.cpp',
    'symbols_with_numbers.cpp',
    'basis_function.cpp',
    'cluster_list.cpp',
    'atoms.cpp',
]

src_files = [cxx_src_folder + x for x in src_files]
src_files.append(cython_folder + 'clease_cxx.pyx')
extra_comp_args = ['-std=c++11']

clease_cxx = Extension("clease_cxx",
                       sources=src_files,
                       include_dirs=[
                           cxx_inc_folder,
                           get_npy_include_folder(), cxx_src_folder,
                           get_python_inc(), cython_folder
                       ],
                       extra_compile_args=extra_comp_args,
                       language="c++")

# Extra packages required with certain aspects of CLEASE, which are not required
# to run the base CLEASE package.
EXTRAS_REQUIRE = {
    # For building the documentation
    'doc': ('sphinx', 'sphinx_rtd_theme'),
    # For running the CLEASE test suite
    'test': (
        'pytest',
        'pytest-mock',
        'mock',
        # Get the histrogram extras, for making nice histogram plots
        # with pytest-benchmark
        'pytest-benchmark[histogram]>=3.4.1',
        'tox>=3.24.0',
    ),
    # Extra nice-to-haves when developing CLEASE
    'dev': (
        'pip',
        'cython',
        'pre-commit',
        'ipython',
        'twine',
        'yapf',
        'pylint',
        'pyclean>=2.0.0',  # For removing __pycache__ and .pyc files
        'pytest-cov',
        'build',
    ),
    'gui': ('clease-gui',),
}
# Make an entry which installs all of the above in one go
# Separate out "gui" from "all"
EXTRAS_REQUIRE['all'] = tuple(
    {package for key, tup in EXTRAS_REQUIRE.items() for package in tup if key != 'gui'})

setup(
    ext_modules=build_ext(clease_cxx),
    extras_require=EXTRAS_REQUIRE,
)



from typing import Dict, Union
from abc import ABC, abstractmethod
from ase import Atoms
from clease.datastructures import SystemChanges
from clease.montecarlo.mc_evaluator import MCEvaluator, construct_evaluator

__all__ = ('BarrierModel', 'BEPBarrier')


class BarrierModel(ABC):
    """
    Base barrier model class
    """

    def __call__(self, system: Union[Atoms, MCEvaluator], system_changes: SystemChanges) -> float:
        """Get the height of the barrier"""
        # Ensure we have a valid evaluator object
        evaluator = construct_evaluator(system)

        return self.evaluate_barrier(evaluator, system_changes)

    @abstractmethod
    def evaluate_barrier(self, evaluator: MCEvaluator, system_changes: SystemChanges) -> float:
        """Evaluate the height of the barrier"""


class BEPBarrier(BarrierModel):
    """
    BEP barrier implements the Bell-Evans-Polanyi model. It is a simple barrier
    model where the barrier is modelled by the dilute barrier, corrected by
    the change in total energy before and after the jump.

    E_a = Q_s + alpha*(E_2 - E_1)

    where E_1 is the total energy of starting configuration and E_2 is the
    total energy of the end configuration. Q_s is a constant that depends
    on the species. If an isolated solute is diffusing in a host material
    E_1 = E_2. Thus, Q_s represents the barrier of an isolated solute. References:

    1. The `Bell-Evans-Polanyi`_ principle.
    2. Andersen, M., Panosetti, C. and Reuter, K., 2019.
       A practical guide to surface kinetic Monte Carlo simulations.
       Frontiers in chemistry, 7, p.202. DOI: https://doi.org/10.3389/fchem.2019.00202

    :param dilute_barrier: Dictionary representing the dilute barrier for
        each species. Example {'Al': 0.05, 'Mg': 0.03}
    :param alpha: Scalar quantity between 0 and 1 that characterises the position
        of the transition state along the reaction coordinate. Default is 0.5.
        Note as alpha is the same for all jumps, it must be viewed as
        controlling the average barrier position.

    .. _Bell-Evans-Polanyi:
        https://en.wikipedia.org/wiki/Bell%E2%80%93Evans%E2%80%93Polanyi_principle>
    """

    def __init__(
        self,
        dilute_barrier: Dict[str, float],
        alpha: float = 0.5,
    ):
        super().__init__()
        self.dilute_barrier = dilute_barrier
        self.alpha = alpha

    def evaluate_barrier(self, evaluator: MCEvaluator, system_changes: SystemChanges) -> float:
        E1 = evaluator.get_energy()
        # Get the energy given a change, remember to undo the changes again.
        E2 = evaluator.get_energy_given_change(system_changes)

        # Extract the jumping species. The system change should involve
        # a vacancy and another element. Thus, the jumping species is the
        # symbol in system_change that is not a vacancy
        old_symb, new_symb = system_changes[0].old_symb, system_changes[0].new_symb

        # Confirm that one of the symbols is a vacancy
        assert old_symb == 'X' or new_symb == 'X'

        jumping_symb = old_symb if old_symb != 'X' else new_symb
        Ea = self.dilute_barrier[jumping_symb] + self.alpha * (E2 - E1)
        return Ea
